@@ -1,23 +1,28 @@
-
 // Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
 // documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
 // rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 // permit persons to whom the Software is furnished to do so.
-//  
+//
 // THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 // WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
 // OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 #include "pyodbc.h"
+#include "wrapper.h"
+#include "textenc.h"
 #include "connection.h"
 #include "cursor.h"
 #include "pyodbcmodule.h"
 #include "errors.h"
-#include "wrapper.h"
 #include "cnxninfo.h"
 #include "sqlwchar.h"
 
+#if PY_MAJOR_VERSION < 3
+static bool IsStringType(PyObject* t) { return (void*)t == (void*)&PyString_Type; }
+static bool IsUnicodeType(PyObject* t) { return (void*)t == (void*)&PyUnicode_Type; }
+#endif
+
 static char connection_doc[] =
     "Connection objects manage connections to the database.\n"
     "\n"
@@ -26,7 +31,7 @@ static char connection_doc[] =
 static Connection* Connection_Validate(PyObject* self)
 {
     Connection* cnxn;
-    
+
     if (self == 0 || !Connection_Check(self))
     {
         PyErr_SetString(PyExc_TypeError, "Connection object required");
@@ -44,7 +49,8 @@ static Connection* Connection_Validate(PyObject* self)
     return cnxn;
 }
 
-static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi, long timeout)
+static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi, long timeout,
+                    Object& encoding)
 {
     // This should have been checked by the global connect function.
     I(PyString_Check(pConnectString) || PyUnicode_Check(pConnectString));
@@ -72,7 +78,7 @@ static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi, long timeou
     if (timeout > 0)
     {
         Py_BEGIN_ALLOW_THREADS
-        ret = SQLSetConnectAttr(hdbc, SQL_ATTR_LOGIN_TIMEOUT, (SQLPOINTER)timeout, SQL_IS_UINTEGER);
+        ret = SQLSetConnectAttr(hdbc, SQL_ATTR_LOGIN_TIMEOUT, (SQLPOINTER)(uintptr_t)timeout, SQL_IS_UINTEGER);
         Py_END_ALLOW_THREADS
         if (!SQL_SUCCEEDED(ret))
             RaiseErrorFromHandle("SQLSetConnectAttr(SQL_ATTR_LOGIN_TIMEOUT)", hdbc, SQL_NULL_HANDLE);
@@ -80,56 +86,24 @@ static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi, long timeou
 
     if (!fAnsi)
     {
-        SQLWChar connectString(pConnectString);
+        // I want to call the W version when possible since the driver can use it as an
+        // indication that we can handle Unicode.  We are going to use the same unicode ending
+        // as we do for binding parameters.
+
+        SQLWChar wchar(pConnectString, SQL_C_WCHAR, encoding, "utf-16le");
+        if (!wchar)
+            return false;
+
         Py_BEGIN_ALLOW_THREADS
-        ret = SQLDriverConnectW(hdbc, 0, connectString.get(), (SQLSMALLINT)connectString.size(), 0, 0, 0, SQL_DRIVER_NOPROMPT);
+        ret = SQLDriverConnectW(hdbc, 0, (SQLWCHAR*)wchar.value(), (SQLSMALLINT)wchar.charlen(), 0, 0, 0, SQL_DRIVER_NOPROMPT);
         Py_END_ALLOW_THREADS
         if (SQL_SUCCEEDED(ret))
             return true;
-
-        // The Unicode function failed.  If the error is that the driver doesn't have a Unicode version (IM001), continue
-        // to the ANSI version.
-        //
-        // I've commented this out since a number of common drivers are returning different errors.  The MySQL 5
-        // driver, for example, returns IM002 "Data source name not found...".
-        //
-        // PyObject* error = GetErrorFromHandle("SQLDriverConnectW", hdbc, SQL_NULL_HANDLE);
-        // if (!HasSqlState(error, "IM001"))
-        // {
-        //     RaiseErrorFromException(error);
-        //     return false;
-        // }
-        // Py_XDECREF(error);
-    }
-        
-    SQLCHAR szConnect[cchMax];
-    if (PyUnicode_Check(pConnectString))
-    {
-        Py_UNICODE* p = PyUnicode_AS_UNICODE(pConnectString);
-        for (Py_ssize_t i = 0, c = PyUnicode_GET_SIZE(pConnectString); i <= c; i++)
-        {
-            if (p[i] > 0xFF)
-            {
-                PyErr_SetString(PyExc_TypeError, "A Unicode connection string was supplied but the driver does "
-                                "not have a Unicode connect function");
-                return false;
-            }
-            szConnect[i] = (SQLCHAR)p[i];
-        }
-    }
-    else
-    {
-#if PY_MAJOR_VERSION < 3
-        const char* p = PyString_AS_STRING(pConnectString);
-        memcpy(szConnect, p, (size_t)(PyString_GET_SIZE(pConnectString) + 1));
-#else
-        PyErr_SetString(PyExc_TypeError, "Connection strings must be Unicode");
-        return false;
-#endif
     }
 
+    SQLWChar ch(pConnectString, SQL_C_CHAR, encoding, "utf-8");
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLDriverConnect(hdbc, 0, szConnect, SQL_NTS, 0, 0, 0, SQL_DRIVER_NOPROMPT);
+    ret = SQLDriverConnect(hdbc, 0, (SQLCHAR*)ch.value(), (SQLSMALLINT)ch.charlen(), 0, 0, 0, SQL_DRIVER_NOPROMPT);
     Py_END_ALLOW_THREADS
     if (SQL_SUCCEEDED(ret))
         return true;
@@ -140,16 +114,14 @@ static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi, long timeou
 }
 
 
-PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi, bool fUnicodeResults, long timeout, bool fReadOnly)
+PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi, long timeout, bool fReadOnly,
+                         PyObject* attrs_before, Object& encoding)
 {
     // pConnectString
     //   A string or unicode object.  (This must be checked by the caller.)
     //
     // fAnsi
     //   If true, do not attempt a Unicode connection.
-    //
-    // fUnicodeResults
-    //   If true, return strings in rows as unicode objects.
 
     //
     // Allocate HDBC and connect
@@ -163,7 +135,44 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
     if (!SQL_SUCCEEDED(ret))
         return RaiseErrorFromHandle("SQLAllocHandle", SQL_NULL_HANDLE, SQL_NULL_HANDLE);
 
-    if (!Connect(pConnectString, hdbc, fAnsi, timeout))
+    //
+    // Attributes that must be set before connecting.
+    //
+
+    if (attrs_before)
+    {
+        Py_ssize_t pos = 0;
+        PyObject* key = 0;
+        PyObject* value = 0;
+        while (PyDict_Next(attrs_before, &pos, &key, &value))
+        {
+            int ikey = 0, ivalue = 0;
+#if PY_MAJOR_VERSION < 3
+            if (PyInt_Check(key))
+                ikey = (int)PyInt_AsLong(key);
+            if (PyInt_Check(value))
+                ivalue = (int)PyInt_AsLong(value);
+#endif
+            if (PyLong_Check(key))
+                ikey = (int)PyLong_AsLong(key);
+            if (PyLong_Check(value))
+                ivalue = (int)PyLong_AsLong(value);
+
+            Py_BEGIN_ALLOW_THREADS
+            ret = SQLSetConnectAttr(hdbc, ikey, (SQLPOINTER)(uintptr_t)ivalue, SQL_IS_INTEGER);
+            Py_END_ALLOW_THREADS
+            if (!SQL_SUCCEEDED(ret))
+            {
+                RaiseErrorFromHandle("SQLSetConnectAttr", hdbc, SQL_NULL_HANDLE);
+                Py_BEGIN_ALLOW_THREADS
+                SQLFreeHandle(SQL_HANDLE_DBC, hdbc);
+                Py_END_ALLOW_THREADS
+                return 0;
+            }
+        }
+    }
+
+    if (!Connect(pConnectString, hdbc, fAnsi, timeout, encoding))
     {
         // Connect has already set an exception.
         Py_BEGIN_ALLOW_THREADS
@@ -173,7 +182,7 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
     }
 
     //
-    // Connected, so allocate the Connection object. 
+    // Connected, so allocate the Connection object.
     //
 
     // Set all variables to something valid, so we don't crash in dealloc if this function fails.
@@ -194,16 +203,59 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
         return 0;
     }
 
-    cnxn->hdbc            = hdbc;
-    cnxn->nAutoCommit     = fAutoCommit ? SQL_AUTOCOMMIT_ON : SQL_AUTOCOMMIT_OFF;
-    cnxn->searchescape    = 0;
-    cnxn->timeout         = 0;
+    cnxn->hdbc         = hdbc;
+    cnxn->nAutoCommit  = fAutoCommit ? SQL_AUTOCOMMIT_ON : SQL_AUTOCOMMIT_OFF;
+    cnxn->searchescape = 0;
+    cnxn->maxwrite     = 0;
+    cnxn->timeout      = 0;
+    cnxn->conv_count   = 0;
+    cnxn->conv_types   = 0;
+    cnxn->conv_funcs   = 0;
+
+    // This is an inefficient default, but should work all the time.  When we are offered
+    // single-byte text we don't actually know what the encoding is.  For example, with SQL
+    // Server the encoding is based on the database's collation.  We ask the driver / DB to
+    // convert to SQL_C_WCHAR and use the ODBC default of UTF-16LE.
+    cnxn->sqlchar_enc.optenc = OPTENC_UTF16LE;
+    cnxn->sqlchar_enc.name   = _strdup("utf-16le");
+    cnxn->sqlchar_enc.ctype  = SQL_C_WCHAR;
+
+    cnxn->sqlwchar_enc.optenc = OPTENC_UTF16LE;
+    cnxn->sqlwchar_enc.name   = _strdup("utf-16le");
+    cnxn->sqlwchar_enc.ctype  = SQL_C_WCHAR;
+
+    cnxn->metadata_enc.optenc = OPTENC_UTF16LE;
+    cnxn->metadata_enc.name   = _strdup("utf-16le");
+    cnxn->metadata_enc.ctype  = SQL_C_WCHAR;
+
+    // Note: I attempted to use UTF-8 here too since it can hold any type, but SQL Server fails
+    // with a data truncation error if we send something encoded in 2 bytes to a column with 1
+    // character.  I don't know if this is a bug in SQL Server's driver or if I'm missing
+    // something, so we'll stay with the default ODBC conversions.
+    cnxn->unicode_enc.optenc = OPTENC_UTF16LE;
+    cnxn->unicode_enc.name   = _strdup("utf-16le");
+    cnxn->unicode_enc.ctype  = SQL_C_WCHAR;
+
 #if PY_MAJOR_VERSION < 3
-    cnxn->unicode_results = fUnicodeResults;
+    cnxn->str_enc.optenc = OPTENC_UTF8;
+    cnxn->str_enc.name   = _strdup("utf-8");
+    cnxn->str_enc.ctype  = SQL_C_CHAR;
+
+    cnxn->sqlchar_enc.to  = TO_UNICODE;
+    cnxn->sqlwchar_enc.to = TO_UNICODE;
+    cnxn->metadata_enc.to = TO_UNICODE;
 #endif
-    cnxn->conv_count      = 0;
-    cnxn->conv_types      = 0;
-    cnxn->conv_funcs      = 0;
+
+    if (!cnxn->sqlchar_enc.name || !cnxn->sqlwchar_enc.name || !cnxn->metadata_enc.name || !cnxn->unicode_enc.name
+#if PY_MAJOR_VERSION < 3
+        || !cnxn->str_enc.name
+#endif
+        )
+    {
+        PyErr_NoMemory();
+        Py_DECREF(cnxn);
+        return 0;
+    }
 
     //
     // Initialize autocommit mode.
@@ -215,7 +267,6 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
 
     if (fAutoCommit == false)
     {
-        SQLRETURN ret;
         Py_BEGIN_ALLOW_THREADS
         ret = SQLSetConnectAttr(cnxn->hdbc, SQL_ATTR_AUTOCOMMIT, (SQLPOINTER)cnxn->nAutoCommit, SQL_IS_UINTEGER);
         Py_END_ALLOW_THREADS
@@ -227,10 +278,9 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
             return 0;
         }
     }
-    
+
     if (fReadOnly)
     {
-        SQLRETURN ret;
         Py_BEGIN_ALLOW_THREADS
         ret = SQLSetConnectAttr(cnxn->hdbc, SQL_ATTR_ACCESS_MODE, (SQLPOINTER)SQL_MODE_READ_ONLY, 0);
         Py_END_ALLOW_THREADS
@@ -262,10 +312,10 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
     cnxn->odbc_minor             = p->odbc_minor;
     cnxn->supports_describeparam = p->supports_describeparam;
     cnxn->datetime_precision     = p->datetime_precision;
+    cnxn->need_long_data_len     = p->need_long_data_len;
     cnxn->varchar_maxlength      = p->varchar_maxlength;
     cnxn->wvarchar_maxlength     = p->wvarchar_maxlength;
     cnxn->binary_maxlength       = p->binary_maxlength;
-    cnxn->need_long_data_len     = p->need_long_data_len;
 
     return reinterpret_cast<PyObject*>(cnxn);
 }
@@ -276,7 +326,7 @@ static void _clear_conv(Connection* cnxn)
     {
         pyodbc_free(cnxn->conv_types);
         cnxn->conv_types = 0;
-        
+
         for (int i = 0; i < cnxn->conv_count; i++)
             Py_XDECREF(cnxn->conv_funcs[i]);
         pyodbc_free(cnxn->conv_funcs);
@@ -286,6 +336,34 @@ static void _clear_conv(Connection* cnxn)
     }
 }
 
+static char set_attr_doc[] =
+    "set_attr(attr_id, value) -> None\n\n"
+    "Calls SQLSetConnectAttr with the given values.\n\n"
+    "attr_id\n"
+    "  The attribute id (integer) to set.  These are ODBC or driver constants.\n\n"
+    "value\n"
+    "  An integer value.\n\n"
+    "At this time, only integer values are supported and are always passed as SQLUINTEGER.";
+
+static PyObject* Connection_set_attr(PyObject* self, PyObject* args)
+{
+    int id;
+    int value;
+    if (!PyArg_ParseTuple(args, "ii", &id, &value))
+        return 0;
+
+    Connection* cnxn = (Connection*)self;
+
+    SQLRETURN ret;
+    Py_BEGIN_ALLOW_THREADS
+    ret = SQLSetConnectAttr(cnxn->hdbc, id, (SQLPOINTER)(intptr_t)value, SQL_IS_INTEGER);
+    Py_END_ALLOW_THREADS
+
+    if (!SQL_SUCCEEDED(ret))
+        return RaiseErrorFromHandle("SQLSetConnectAttr", cnxn->hdbc, SQL_NULL_HANDLE);
+    Py_RETURN_NONE;
+}
+
 static char conv_clear_doc[] =
     "clear_output_converters() --> None\n\n"
     "Remove all output converter functions.";
@@ -293,7 +371,7 @@ static char conv_clear_doc[] =
 static PyObject* Connection_conv_clear(PyObject* self, PyObject* args)
 {
     UNUSED(args);
-    
+
     Connection* cnxn = (Connection*)self;
     _clear_conv(cnxn);
     Py_RETURN_NONE;
@@ -308,24 +386,35 @@ static int Connection_clear(PyObject* self)
 
     if (cnxn->hdbc != SQL_NULL_HANDLE)
     {
-        // REVIEW: Release threads? (But make sure you zero out hdbc *first*!
-
         TRACE("cnxn.clear cnxn=%p hdbc=%d\n", cnxn, cnxn->hdbc);
 
+        HDBC hdbc = cnxn->hdbc;
+        cnxn->hdbc = SQL_NULL_HANDLE;
         Py_BEGIN_ALLOW_THREADS
         if (cnxn->nAutoCommit == SQL_AUTOCOMMIT_OFF)
-            SQLEndTran(SQL_HANDLE_DBC, cnxn->hdbc, SQL_ROLLBACK);
+            SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_ROLLBACK);
 
-        SQLDisconnect(cnxn->hdbc);
-        SQLFreeHandle(SQL_HANDLE_DBC, cnxn->hdbc);
+        SQLDisconnect(hdbc);
+        SQLFreeHandle(SQL_HANDLE_DBC, hdbc);
         Py_END_ALLOW_THREADS
-        
-        cnxn->hdbc = SQL_NULL_HANDLE;
     }
 
     Py_XDECREF(cnxn->searchescape);
     cnxn->searchescape = 0;
-    
+
+    free((void*)cnxn->sqlchar_enc.name);
+    cnxn->sqlchar_enc.name = 0;
+    free((void*)cnxn->sqlwchar_enc.name);
+    cnxn->sqlwchar_enc.name = 0;
+    free((void*)cnxn->metadata_enc.name);
+    cnxn->metadata_enc.name = 0;
+    free((void*)cnxn->unicode_enc.name);
+    cnxn->unicode_enc.name = 0;
+#if PY_MAJOR_VERSION < 3
+    free((void*)cnxn->str_enc.name);
+    cnxn->str_enc.name = 0;
+#endif
+
     _clear_conv(cnxn);
 
     return 0;
@@ -346,11 +435,11 @@ static char close_doc[] =
     "\n"
     "Note that closing a connection without committing the changes first will cause\n"
     "an implicit rollback to be performed.";
-    
+
 static PyObject* Connection_close(PyObject* self, PyObject* args)
 {
     UNUSED(args);
-    
+
     Connection* cnxn = Connection_Validate(self);
     if (!cnxn)
         return 0;
@@ -363,7 +452,7 @@ static PyObject* Connection_close(PyObject* self, PyObject* args)
 static PyObject* Connection_cursor(PyObject* self, PyObject* args)
 {
     UNUSED(args);
-    
+
     Connection* cnxn = Connection_Validate(self);
     if (!cnxn)
         return 0;
@@ -551,6 +640,7 @@ static const GetInfoType aInfoTypes[] = {
     { SQL_XOPEN_CLI_YEAR, GI_STRING },
 };
 
+
 static PyObject* Connection_getinfo(PyObject* self, PyObject* args)
 {
     Connection* cnxn = Connection_Validate(self);
@@ -569,7 +659,7 @@ static PyObject* Connection_getinfo(PyObject* self, PyObject* args)
     }
 
     if (i == _countof(aInfoTypes))
-        return RaiseErrorV(0, ProgrammingError, "Invalid getinfo value: %d", infotype);
+        return RaiseErrorV(0, ProgrammingError, "Unsupported getinfo value: %d", infotype);
 
     char szBuffer[0x1000];
     SQLSMALLINT cch = 0;
@@ -592,11 +682,11 @@ static PyObject* Connection_getinfo(PyObject* self, PyObject* args)
         result = (szBuffer[0] == 'Y') ? Py_True : Py_False;
         Py_INCREF(result);
         break;
-            
+
     case GI_STRING:
         result = PyString_FromStringAndSize(szBuffer, (Py_ssize_t)cch);
         break;
-        
+
     case GI_UINTEGER:
     {
         SQLUINTEGER n = *(SQLUINTEGER*)szBuffer; // Does this work on PPC or do we need a union?
@@ -610,7 +700,7 @@ static PyObject* Connection_getinfo(PyObject* self, PyObject* args)
 #endif
         break;
     }
-    
+
     case GI_USMALLINT:
         result = PyInt_FromLong(*(SQLUSMALLINT*)szBuffer);
         break;
@@ -643,11 +733,11 @@ PyObject* Connection_endtrans(Connection* cnxn, SQLSMALLINT type)
 static PyObject* Connection_commit(PyObject* self, PyObject* args)
 {
     UNUSED(args);
-    
+
     Connection* cnxn = Connection_Validate(self);
     if (!cnxn)
         return 0;
-    
+
     TRACE("commit: cnxn=%p hdbc=%d\n", cnxn, cnxn->hdbc);
 
     return Connection_endtrans(cnxn, SQL_COMMIT);
@@ -656,19 +746,19 @@ static PyObject* Connection_commit(PyObject* self, PyObject* args)
 static PyObject* Connection_rollback(PyObject* self, PyObject* args)
 {
     UNUSED(args);
-    
+
     Connection* cnxn = Connection_Validate(self);
     if (!cnxn)
         return 0;
-    
+
     TRACE("rollback: cnxn=%p hdbc=%d\n", cnxn, cnxn->hdbc);
 
     return Connection_endtrans(cnxn, SQL_ROLLBACK);
 }
 
-static char cursor_doc[] = 
+static char cursor_doc[] =
     "Return a new Cursor object using the connection.";
-    
+
 static char execute_doc[] =
     "execute(sql, [params]) --> Cursor\n"
     "\n"
@@ -717,7 +807,7 @@ static int Connection_setautocommit(PyObject* self, PyObject* value, void* closu
         PyErr_SetString(PyExc_TypeError, "Cannot delete the autocommit attribute.");
         return -1;
     }
-    
+
     uintptr_t nAutoCommit = PyObject_IsTrue(value) ? SQL_AUTOCOMMIT_ON : SQL_AUTOCOMMIT_OFF;
     SQLRETURN ret;
     Py_BEGIN_ALLOW_THREADS
@@ -738,7 +828,7 @@ static int Connection_setautocommit(PyObject* self, PyObject* value, void* closu
 static PyObject* Connection_getsearchescape(PyObject* self, void* closure)
 {
     UNUSED(closure);
-    
+
     Connection* cnxn = (Connection*)self;
 
     if (!cnxn->searchescape)
@@ -760,6 +850,47 @@ static PyObject* Connection_getsearchescape(PyObject* self, void* closure)
     return cnxn->searchescape;
 }
 
+static PyObject* Connection_getmaxwrite(PyObject* self, void* closure)
+{
+    UNUSED(closure);
+
+    Connection* cnxn = Connection_Validate(self);
+    if (!cnxn)
+        return 0;
+
+    return PyLong_FromSsize_t(cnxn->maxwrite);
+}
+
+static int Connection_setmaxwrite(PyObject* self, PyObject* value, void* closure)
+{
+    UNUSED(closure);
+
+    Connection* cnxn = Connection_Validate(self);
+    if (!cnxn)
+        return -1;
+
+    if (value == 0)
+    {
+        PyErr_SetString(PyExc_TypeError, "Cannot delete the maxwrite attribute.");
+        return -1;
+    }
+    long maxwrite = PyLong_AsLong(value);
+    if (PyErr_Occurred())
+        return -1;
+
+    Py_ssize_t minval = 255;
+
+    if (maxwrite != 0 && maxwrite < minval)
+    {
+        PyErr_Format(PyExc_ValueError, "Cannot set maxwrite less than %d unless setting to 0.", (int)minval);
+        return -1;
+    }
+
+    cnxn->maxwrite = maxwrite;
+
+    return 0;
+}
+
 
 static PyObject* Connection_gettimeout(PyObject* self, void* closure)
 {
@@ -796,7 +927,7 @@ static int Connection_settimeout(PyObject* self, PyObject* value, void* closure)
 
     SQLRETURN ret;
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLSetConnectAttr(cnxn->hdbc, SQL_ATTR_CONNECTION_TIMEOUT, (SQLPOINTER)timeout, SQL_IS_UINTEGER);
+    ret = SQLSetConnectAttr(cnxn->hdbc, SQL_ATTR_CONNECTION_TIMEOUT, (SQLPOINTER)(uintptr_t)timeout, SQL_IS_UINTEGER);
     Py_END_ALLOW_THREADS
     if (!SQL_SUCCEEDED(ret))
     {
@@ -827,7 +958,7 @@ static bool _add_converter(PyObject* self, SQLSMALLINT sqltype, PyObject* func)
             }
         }
     }
-    
+
     int          oldcount = cnxn->conv_count;
     SQLSMALLINT* oldtypes = cnxn->conv_types;
     PyObject**   oldfuncs = cnxn->conv_funcs;
@@ -857,7 +988,7 @@ static bool _add_converter(PyObject* self, SQLSMALLINT sqltype, PyObject* func)
     if (oldcount != 0)
     {
         // copy old items
-        memcpy(&newtypes[1], oldtypes, sizeof(int) * oldcount);
+        memcpy(&newtypes[1], oldtypes, sizeof(SQLSMALLINT) * oldcount);
         memcpy(&newfuncs[1], oldfuncs, sizeof(PyObject*) * oldcount);
 
         pyodbc_free(oldtypes);
@@ -881,8 +1012,13 @@ static char conv_add_doc[] =
     "func\n"
     "  The converter function which will be called with a single parameter, the\n"
     "  value, and should return the converted value.  If the value is NULL, the\n"
-    "  parameter will be None.  Otherwise it will be a Python string.";
-
+    "  parameter will be None.  Otherwise it will be a "
+#if PY_MAJOR_VERSION >= 3
+    "bytes object."
+#else
+    "str object with the raw bytes."
+#endif
+    ;
 
 static PyObject* Connection_conv_add(PyObject* self, PyObject* args)
 {
@@ -897,6 +1033,266 @@ static PyObject* Connection_conv_add(PyObject* self, PyObject* args)
     Py_RETURN_NONE;
 }
 
+static void NormalizeCodecName(const char* src, char* dest, size_t cbDest)
+{
+    // Copies the codec name to dest, lowercasing it and replacing underscores with dashes.
+    // (Same as _Py_normalize_encoding which is not public.)  It also wraps the value with
+    // pipes so we can search with it.
+    //
+    // UTF_8 --> |utf-8|
+    //
+    // This is an internal function - it will truncate so you should use a buffer bigger than
+    // anything you expect to search for.
+
+    char* pch = &dest[0];
+    char* pchLast = pch + cbDest - 2; // -2 -> leave room for pipe and null
+
+    *pch++ = '|';
+
+    while (*src && pch < pchLast)
+    {
+        if (isupper(*src))
+        {
+            *pch++ = (char)tolower(*src++);
+        }
+        else if (*src == '_')
+        {
+            *pch++ = '-';
+            src++;
+        }
+        else
+        {
+            *pch++ = *src++;
+        }
+    }
+
+    *pch++ = '|';
+    *pch = '\0';
+}
+
+static bool SetTextEncCommon(TextEnc& enc, const char* encoding, int ctype, bool allow_raw)
+{
+    // Code common to setencoding and setdecoding.
+
+    if (!encoding)
+    {
+        PyErr_Format(PyExc_ValueError, "encoding is required");
+        return false;
+    }
+
+    // Normalize the names so we don't have to worry about case or dashes vs underscores.
+    // We'll lowercase everything and convert underscores to dashes.  The results are then
+    // surrounded with pipes so we can search strings.  (See the `strstr` calls below.)
+    char lower[30];
+    NormalizeCodecName(encoding, lower, sizeof(lower));
+
+#if PY_MAJOR_VERSION < 3
+    if (strcmp(lower, "|raw|") == 0)
+    {
+        if (!allow_raw)
+        {
+            // Give a better error message for 'raw' than "not a registered codec".  It is never
+            // registered.
+            PyErr_Format(PyExc_ValueError, "Raw codec is only allowed for str / SQL_CHAR");
+            return false;
+        }
+    }
+    else if (!PyCodec_KnownEncoding(encoding))
+    {
+        PyErr_Format(PyExc_ValueError, "not a registered codec: '%s'", encoding);
+        return false;
+    }
+#else
+    if (!PyCodec_KnownEncoding(encoding))
+    {
+        PyErr_Format(PyExc_ValueError, "not a registered codec: '%s'", encoding);
+        return false;
+    }
+#endif
+
+    if (ctype != 0 && ctype != SQL_WCHAR && ctype != SQL_CHAR)
+    {
+        PyErr_Format(PyExc_ValueError, "Invalid ctype %d.  Must be SQL_CHAR or SQL_WCHAR", ctype);
+        return false;
+    }
+
+    char* cpy = _strdup(encoding);
+    if (!cpy)
+    {
+        PyErr_NoMemory();
+        return false;
+    }
+
+    free((void*)enc.name);
+    enc.name = cpy;
+
+    if (strstr("|utf-8|utf8|", lower))
+    {
+        enc.optenc = OPTENC_UTF8;
+        enc.ctype  = (SQLSMALLINT)(ctype ? ctype : SQL_C_CHAR);
+    }
+    else if (strstr("|utf-16|utf16|", lower))
+    {
+        enc.optenc = OPTENC_UTF16;
+        enc.ctype  = (SQLSMALLINT)(ctype ? ctype : SQL_C_WCHAR);
+    }
+    else if (strstr("|utf-16-be|utf-16be|utf16be|", lower))
+    {
+        enc.optenc = OPTENC_UTF16BE;
+        enc.ctype  = (SQLSMALLINT)(ctype ? ctype : SQL_C_WCHAR);
+    }
+    else if (strstr("|utf-16-le|utf-16le|utf16le|", lower))
+    {
+        enc.optenc = OPTENC_UTF16LE;
+        enc.ctype  = (SQLSMALLINT)(ctype ? ctype : SQL_C_WCHAR);
+    }
+    else if (strstr("|utf-32|utf32|", lower))
+    {
+        enc.optenc = OPTENC_UTF32;
+        enc.ctype  = (SQLSMALLINT)(ctype ? ctype : SQL_C_WCHAR);
+    }
+    else if (strstr("|utf-32-be|utf-32be|utf32be|", lower))
+    {
+        enc.optenc = OPTENC_UTF32BE;
+        enc.ctype  = (SQLSMALLINT)(ctype ? ctype : SQL_C_WCHAR);
+    }
+    else if (strstr("|utf-32-le|utf-32le|utf32le|", lower))
+    {
+        enc.optenc = OPTENC_UTF32LE;
+        enc.ctype  = (SQLSMALLINT)(ctype ? ctype : SQL_C_WCHAR);
+    }
+    else if (strstr("|latin-1|latin1|iso-8859-1|iso8859-1|", lower))
+    {
+        enc.optenc = OPTENC_LATIN1;
+        enc.ctype  = (SQLSMALLINT)(ctype ? ctype : SQL_C_CHAR);
+    }
+#if PY_MAJOR_VERSION < 3
+    else if (strstr("|raw|", lower))
+    {
+        enc.optenc = OPTENC_RAW;
+        enc.ctype  = SQL_C_CHAR;
+    }
+#endif
+    else
+    {
+        enc.optenc = OPTENC_NONE;
+        enc.ctype  = SQL_C_CHAR;
+    }
+
+    return true;
+}
+
+static PyObject* Connection_setencoding(PyObject* self, PyObject* args, PyObject* kwargs)
+{
+    Connection* cnxn = (Connection*)self;
+
+#if PY_MAJOR_VERSION >= 3
+    // In Python 3 we only support encodings for Unicode text.
+    char* encoding = 0;
+    int ctype = 0;
+    static char *kwlist[] = { "encoding", "ctype", 0 };
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|si", kwlist, &encoding, &ctype))
+        return 0;
+    TextEnc& enc = cnxn->unicode_enc;
+    bool allow_raw = false;
+#else
+    // In Python 2, we support encodings for Unicode and strings.
+    PyObject* from_type;
+    char* encoding = 0;
+    int ctype = 0;
+    static char *kwlist[] = { "fromtype", "encoding", "ctype", 0 };
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|si", kwlist, &from_type, &encoding, &ctype))
+        return 0;
+
+    if (!IsUnicodeType(from_type) && ! IsStringType(from_type))
+        return PyErr_Format(PyExc_TypeError, "fromtype must be str or unicode");
+
+    TextEnc& enc = IsStringType(from_type) ? cnxn->str_enc : cnxn->unicode_enc;
+    bool allow_raw = IsStringType(from_type);
+#endif
+
+    if (!SetTextEncCommon(enc, encoding, ctype, allow_raw))
+        return 0;
+
+    Py_RETURN_NONE;
+}
+
+static char setdecoding_doc[] =
+#if PY_MAJOR_VERSION >= 3
+    "setdecoding(sqltype, encoding=None, ctype=None) --> None\n"
+#else
+    "setdecoding(sqltype, encoding=None, ctype=None, to=None) --> None\n"
+#endif
+    "\n"
+    "Configures how text of type `ctype` (SQL_CHAR or SQL_WCHAR) is decoded\n"
+    "when read from the database.\n"
+    "\n"
+    "When reading, the database will assign one of the sqltypes to text columns.\n"
+    "pyodbc uses this lookup the decoding information set by this function.\n"
+    "sqltype: pyodbc.SQL_CHAR or pyodbc.SQL_WCHAR\n\n"
+    "encoding: A registered Python encoding such as \"utf-8\".\n\n"
+#if PY_MAJOR_VERSION < 3
+    "to: the desired Python object type - str or unicode"
+#endif
+    "ctype: The C data type should be requested.  Set this to SQL_CHAR for\n"
+    "  single-byte encodings like UTF-8 and to SQL_WCHAR for two-byte encodings\n"
+    "  like UTF-16.";
+
+
+static PyObject* Connection_setdecoding(PyObject* self, PyObject* args, PyObject* kwargs)
+{
+    Connection* cnxn = (Connection*)self;
+
+    int sqltype;
+    char* encoding = 0;
+    int ctype = 0;
+    bool allow_raw = false;
+
+#if PY_MAJOR_VERSION >= 3
+    static char *kwlist[] = {"sqltype", "encoding", "ctype", NULL};
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i|si", kwlist, &sqltype, &encoding, &ctype))
+        return 0;
+#else
+    int to = 0;
+    PyObject* toObj = 0;
+    static char *kwlist[] = {"sqltype", "encoding", "ctype", "to", NULL};
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i|siO", kwlist, &sqltype, &encoding, &ctype, &toObj))
+        return 0;
+
+    if (toObj)
+    {
+        if (IsUnicodeType(toObj))
+            to = TO_UNICODE;
+        else if (IsStringType(toObj))
+            to = TO_STR;
+        else
+            return PyErr_Format(PyExc_ValueError, "`to` can only be unicode or str");
+    }
+
+    allow_raw = (sqltype == SQL_CHAR && to != TO_UNICODE);
+#endif
+
+    if (sqltype != SQL_WCHAR && sqltype != SQL_CHAR && sqltype != SQL_WMETADATA)
+        return PyErr_Format(PyExc_ValueError, "Invalid sqltype %d.  Must be SQL_CHAR or SQL_WCHAR or SQL_WMETADATA", sqltype);
+
+    TextEnc& enc = (sqltype == SQL_CHAR) ? cnxn->sqlchar_enc :
+        ((sqltype == SQL_WMETADATA) ? cnxn->metadata_enc : cnxn->sqlwchar_enc);
+
+    if (!SetTextEncCommon(enc, encoding, ctype, allow_raw))
+        return 0;
+
+#if PY_MAJOR_VERSION < 3
+    if (!to && enc.optenc == OPTENC_RAW)
+        enc.to = TO_STR;
+    else
+        enc.to = to ? to : TO_UNICODE;
+#endif
+
+    Py_RETURN_NONE;
+}
+
+
+
 static char enter_doc[] = "__enter__() -> self.";
 static PyObject* Connection_enter(PyObject* self, PyObject* args)
 {
@@ -913,17 +1309,21 @@ static PyObject* Connection_exit(PyObject* self, PyObject* args)
     // If an error has occurred, `args` will be a tuple of 3 values.  Otherwise it will be a tuple of 3 `None`s.
     I(PyTuple_Check(args));
 
-    if (cnxn->nAutoCommit == SQL_AUTOCOMMIT_OFF && PyTuple_GetItem(args, 0) == Py_None)
+    if (cnxn->nAutoCommit == SQL_AUTOCOMMIT_OFF)
     {
+        SQLSMALLINT CompletionType = (PyTuple_GetItem(args, 0) == Py_None) ? SQL_COMMIT : SQL_ROLLBACK;
         SQLRETURN ret;
         Py_BEGIN_ALLOW_THREADS
-        ret = SQLEndTran(SQL_HANDLE_DBC, cnxn->hdbc, SQL_COMMIT);
+        ret = SQLEndTran(SQL_HANDLE_DBC, cnxn->hdbc, CompletionType);
         Py_END_ALLOW_THREADS
 
         if (!SQL_SUCCEEDED(ret))
-            return RaiseErrorFromHandle("SQLEndTran(SQL_COMMIT)", cnxn->hdbc, SQL_NULL_HANDLE);
+        {
+            const char* szFunc = (CompletionType == SQL_COMMIT) ? "SQLEndTran(SQL_COMMIT)" : "SQLEndTran(SQL_ROLLBACK)";
+            return RaiseErrorFromHandle(szFunc, cnxn->hdbc, SQL_NULL_HANDLE);
+        }
     }
-    
+
     Py_RETURN_NONE;
 }
 
@@ -938,9 +1338,12 @@ static struct PyMethodDef Connection_methods[] =
     { "getinfo",                 Connection_getinfo,         METH_VARARGS, getinfo_doc    },
     { "add_output_converter",    Connection_conv_add,        METH_VARARGS, conv_add_doc   },
     { "clear_output_converters", Connection_conv_clear,      METH_NOARGS,  conv_clear_doc },
+    { "setdecoding",             (PyCFunction)Connection_setdecoding,     METH_VARARGS|METH_KEYWORDS, setdecoding_doc },
+    { "setencoding",             (PyCFunction)Connection_setencoding,     METH_VARARGS|METH_KEYWORDS, 0 },
+    { "set_attr",                Connection_set_attr,        METH_VARARGS, set_attr_doc   },
     { "__enter__",               Connection_enter,           METH_NOARGS,  enter_doc      },
     { "__exit__",                Connection_exit,            METH_VARARGS, exit_doc       },
-    
+
     { 0, 0, 0, 0 }
 };
 
@@ -952,6 +1355,7 @@ static PyGetSetDef Connection_getseters[] = {
       "Returns True if the connection is in autocommit mode; False otherwise.", 0 },
     { "timeout", Connection_gettimeout, Connection_settimeout,
       "The timeout in seconds, zero means no timeout.", 0 },
+    { "maxwrite", Connection_getmaxwrite, Connection_setmaxwrite, "The maximum bytes to write before using SQLPutData.", 0 },
     { 0 }
 };
 
