@@ -16,6 +16,8 @@ struct Cursor;
 
 extern PyTypeObject ConnectionType;
 
+struct TextEnc;
+
 struct Connection
 {
     PyObject_HEAD
@@ -40,25 +42,48 @@ struct Connection
     // The column size of datetime columns, obtained from SQLGetInfo(), used to determine the datetime precision.
     int datetime_precision;
 
-#if PY_MAJOR_VERSION < 3
-    bool unicode_results;
-    // If true, ANSI columns are returned as Unicode.
-#endif
-
     // The connection timeout in seconds.
     long timeout;
 
-    PyObject* unicode_encoding;
-    // The optional Unicode encoding of the database.  Unicode strings are
-    // encoded when sent and decoded when received.
-    //
-    // If not provided, UCS-2 is used.
+    TextEnc sqlchar_enc;        // encoding used when reading SQL_CHAR data
+    TextEnc sqlwchar_enc;       // encoding used when reading SQL_WCHAR data
+    TextEnc unicode_enc;        // encoding used when writing unicode strings
+#if PY_MAJOR_VERSION < 3
+    TextEnc str_enc;            // encoding used when writing non-unicode strings
+#endif
+
+    TextEnc metadata_enc;
+    // Used when reading column names for Cursor.description.  I originally thought I could use
+    // the TextEncs above based on whether I called SQLDescribeCol vs SQLDescribeColW.
+    // Unfortunately it looks like PostgreSQL and MySQL (and probably others) ignore the ODBC
+    // specification regarding encoding everywhere *except* in these functions - SQLDescribeCol
+    // seems to always return UTF-16LE by them regardless of the connection settings.
+
+    long maxwrite;
+    // Used to override varchar_maxlength, etc.  Those are initialized from
+    // SQLGetTypeInfo but some drivers (e.g. psqlodbc) return almost arbitrary
+    // values (like 255 chars) leading to very slow insert performance (lots of
+    // small calls to SQLPutData).  If this is zero the values from
+    // SQLGetTypeInfo are used.  Otherwise this value is used.
 
     // These are copied from cnxn info for performance and convenience.
 
     int varchar_maxlength;
     int wvarchar_maxlength;
     int binary_maxlength;
+
+    SQLLEN GetMaxLength(SQLSMALLINT ctype) const
+    {
+        I(ctype == SQL_C_BINARY || ctype == SQL_C_WCHAR || ctype == SQL_C_CHAR);
+        if (maxwrite != 0)
+            return maxwrite;
+        if (ctype == SQL_C_BINARY)
+            return binary_maxlength;
+        if (ctype == SQL_C_WCHAR)
+            return wvarchar_maxlength;
+        return varchar_maxlength;
+    }
+
     bool need_long_data_len;
 
     // Output conversions.  Maps from SQL type in conv_types to the converter function in conv_funcs.
@@ -80,7 +105,8 @@ struct Connection
  * Used by the module's connect function to create new connection objects.  If unable to connect to the database, an
  * exception is set and zero is returned.
  */
-PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi, bool fUnicodeResults, long timeout, bool fReadOnly);
+PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi, long timeout, bool fReadOnly,
+                         PyObject* attrs_before, Object& encoding);
 
 /*
  * Used by the Cursor to implement commit and rollback.
