@@ -1,4 +1,3 @@
-
 // Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
 // documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
 // rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
@@ -14,6 +13,8 @@
 // actually used.
 
 #include "pyodbc.h"
+#include "wrapper.h"
+#include "textenc.h"
 #include "cursor.h"
 #include "pyodbcmodule.h"
 #include "connection.h"
@@ -25,7 +26,6 @@
 #include "dbspecific.h"
 #include "sqlwchar.h"
 #include <datetime.h>
-#include "wrapper.h"
 
 enum
 {
@@ -126,106 +126,6 @@ inline bool IsNumericType(SQLSMALLINT sqltype)
 }
 
 
-static PyObject* PythonTypeFromSqlType(Cursor* cur, const SQLCHAR* name, SQLSMALLINT type)
-{
-    // Returns a type object ('int', 'str', etc.) for the given ODBC C type.  This is used to populate
-    // Cursor.description with the type of Python object that will be returned for each column.
-    //
-    // name
-    //   The name of the column, only used to create error messages.
-    //
-    // type
-    //   The ODBC C type (SQL_C_CHAR, etc.) of the column.
-    //
-    // The returned object does not have its reference count incremented!
-
-    int conv_index = GetUserConvIndex(cur, type);
-    if (conv_index != -1)
-        return (PyObject*)&PyString_Type;
-
-    PyObject* pytype = 0;
-
-    switch (type)
-    {
-    case SQL_CHAR:
-    case SQL_VARCHAR:
-    case SQL_LONGVARCHAR:
-    case SQL_GUID:
-    case SQL_SS_XML:
-#if PY_MAJOR_VERSION < 3
-        if (cur->cnxn->unicode_results)
-            pytype = (PyObject*)&PyUnicode_Type;
-        else
-            pytype = (PyObject*)&PyString_Type;
-#else
-        pytype = (PyObject*)&PyString_Type;
-#endif
-        break;
-
-    case SQL_DECIMAL:
-    case SQL_NUMERIC:
-        pytype = (PyObject*)decimal_type;
-        break;
-
-    case SQL_REAL:
-    case SQL_FLOAT:
-    case SQL_DOUBLE:
-        pytype = (PyObject*)&PyFloat_Type;
-        break;
-
-    case SQL_SMALLINT:
-    case SQL_INTEGER:
-    case SQL_TINYINT:
-        pytype = (PyObject*)&PyInt_Type;
-        break;
-
-    case SQL_TYPE_DATE:
-        pytype = (PyObject*)PyDateTimeAPI->DateType;
-        break;
-
-    case SQL_TYPE_TIME:
-    case SQL_SS_TIME2:          // SQL Server 2008+
-        pytype = (PyObject*)PyDateTimeAPI->TimeType;
-        break;
-
-    case SQL_TYPE_TIMESTAMP:
-        pytype = (PyObject*)PyDateTimeAPI->DateTimeType;
-        break;
-
-    case SQL_BIGINT:
-        pytype = (PyObject*)&PyLong_Type;
-        break;
-
-    case SQL_BIT:
-        pytype = (PyObject*)&PyBool_Type;
-        break;
-
-    case SQL_BINARY:
-    case SQL_VARBINARY:
-    case SQL_LONGVARBINARY:
-#if PY_MAJOR_VERSION >= 3
-        pytype = (PyObject*)&PyBytes_Type;
-#else
-        pytype = (PyObject*)&PyBuffer_Type;
-#endif
-        break;
-
-
-    case SQL_WCHAR:
-    case SQL_WVARCHAR:
-    case SQL_WLONGVARCHAR:
-        pytype = (PyObject*)&PyUnicode_Type;
-        break;
-
-    default:
-        return RaiseErrorV(0, 0, "ODBC data type %d is not supported.  Cannot read column %s.", type, (const char*)name);
-    }
-
-    Py_INCREF(pytype);
-    return pytype;
-}
-
-
 static bool create_name_map(Cursor* cur, SQLSMALLINT field_count, bool lower)
 {
     // Called after an execute to construct the map shared by rows.
@@ -254,14 +154,15 @@ static bool create_name_map(Cursor* cur, SQLSMALLINT field_count, bool lower)
 
     for (int i = 0; i < field_count; i++)
     {
-        SQLCHAR name[300];
+        ODBCCHAR szName[300];
+        SQLSMALLINT cchName;
         SQLSMALLINT nDataType;
         SQLULEN nColSize;           // precision
         SQLSMALLINT cDecimalDigits; // scale
         SQLSMALLINT nullable;
 
         Py_BEGIN_ALLOW_THREADS
-        ret = SQLDescribeCol(cur->hstmt, (SQLUSMALLINT)(i + 1), name, _countof(name), 0, &nDataType, &nColSize, &cDecimalDigits, &nullable);
+        ret = SQLDescribeColW(cur->hstmt, (SQLUSMALLINT)(i + 1), (SQLWCHAR*)szName, _countof(szName), &cchName, &nDataType, &nColSize, &cDecimalDigits, &nullable);
         Py_END_ALLOW_THREADS
 
         if (cur->cnxn->hdbc == SQL_NULL_HANDLE)
@@ -277,12 +178,41 @@ static bool create_name_map(Cursor* cur, SQLSMALLINT field_count, bool lower)
             goto done;
         }
 
-        TRACE("Col %d: type=%d colsize=%d\n", (i+1), (int)nDataType, (int)nColSize);
+        const TextEnc& enc = cur->cnxn->metadata_enc;
+
+        // HACK: I don't know the exact issue, but iODBC + Teradata results in either UCS4 data
+        // or 4-byte SQLWCHAR.  I'm going to use UTF-32 as an indication that's what we have.
+
+        Py_ssize_t cbName = cchName;
+        switch (enc.optenc)
+        {
+        case OPTENC_UTF32:
+        case OPTENC_UTF32LE:
+        case OPTENC_UTF32BE:
+            cbName *= 4;
+            break;
+        default:
+            if (enc.ctype == SQL_C_WCHAR)
+                cbName *= 2;
+            break;
+        }
+
+        TRACE("Col %d: type=%s (%d) colsize=%d\n", (i+1), SqlTypeName(nDataType), (int)nDataType, (int)nColSize);
+
+        Object name(TextBufferToObject(enc, szName, cbName));
+
+        if (!name)
+            goto done;
 
         if (lower)
-            _strlwr((char*)name);
+        {
+            PyObject* l = PyObject_CallMethod(name, "lower", 0);
+            if (!l)
+                goto done;
+            name.Attach(l);
+        }
 
-        type = PythonTypeFromSqlType(cur, name, nDataType);
+        type = PythonTypeFromSqlType(cur, nDataType);
         if (!type)
             goto done;
 
@@ -321,8 +251,8 @@ static bool create_name_map(Cursor* cur, SQLSMALLINT field_count, bool lower)
             }
         }
 
-        colinfo = Py_BuildValue("(sOOiiiO)",
-                                (char*)name,
+        colinfo = Py_BuildValue("(OOOiiiO)",
+                                name.Get(),
                                 type,                // type_code
                                 Py_None,             // display size
                                 (int)nColSize,       // internal_size
@@ -332,14 +262,13 @@ static bool create_name_map(Cursor* cur, SQLSMALLINT field_count, bool lower)
         if (!colinfo)
             goto done;
 
-
         nullable_obj = 0;
 
         index = PyInt_FromLong(i);
         if (!index)
             goto done;
 
-        PyDict_SetItemString(colmap, (const char*)name, index);
+        PyDict_SetItem(colmap, name.Get(), index);
         Py_DECREF(index);       // SetItemString increments
         index = 0;
 
@@ -458,11 +387,16 @@ static void closeimpl(Cursor* cur)
     {
         HSTMT hstmt = cur->hstmt;
         cur->hstmt = SQL_NULL_HANDLE;
-        Py_BEGIN_ALLOW_THREADS
-        SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
-        Py_END_ALLOW_THREADS
-    }
 
+        SQLRETURN ret;
+        Py_BEGIN_ALLOW_THREADS
+        ret = SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
+        Py_END_ALLOW_THREADS
+
+        // If there is already an exception, don't overwrite it.
+        if (!SQL_SUCCEEDED(ret) && !PyErr_Occurred())
+            RaiseErrorFromHandle("SQLFreeHandle", cur->cnxn->hdbc, SQL_NULL_HANDLE);
+    }
 
     Py_XDECREF(cur->pPreparedSQL);
     Py_XDECREF(cur->description);
@@ -490,6 +424,9 @@ static PyObject* Cursor_close(PyObject* self, PyObject* args)
 
     closeimpl(cursor);
 
+    if (PyErr_Occurred())
+        return 0;
+
     Py_INCREF(Py_None);
     return Py_None;
 }
@@ -544,7 +481,6 @@ bool InitColumnInfo(Cursor* cursor, SQLUSMALLINT iCol, ColumnInfo* pinfo)
         RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
         return false;
     }
-
     if (!SQL_SUCCEEDED(ret))
     {
         RaiseErrorFromHandle("SQLDescribeCol", cursor->cnxn->hdbc, cursor->hstmt);
@@ -677,23 +613,35 @@ static PyObject* execute(Cursor* cur, PyObject* pSql, PyObject* params, bool ski
         cur->pPreparedSQL = 0;
 
         szLastFunction = "SQLExecDirect";
+
+        const TextEnc* penc = 0;
+
 #if PY_MAJOR_VERSION < 3
         if (PyString_Check(pSql))
         {
-            Py_BEGIN_ALLOW_THREADS
-            ret = SQLExecDirect(cur->hstmt, (SQLCHAR*)PyString_AS_STRING(pSql), SQL_NTS);
-            Py_END_ALLOW_THREADS
+            penc = &cur->cnxn->str_enc;
         }
         else
 #endif
         {
-            SQLWChar query(pSql);
-            if (!query)
-                return 0;
-            Py_BEGIN_ALLOW_THREADS
-            ret = SQLExecDirectW(cur->hstmt, query.get(), SQL_NTS);
-            Py_END_ALLOW_THREADS
+            penc = &cur->cnxn->unicode_enc;
         }
+
+        Object query(penc->Encode(pSql));
+        if (!query)
+            return 0;
+
+        bool isWide = (penc->ctype == SQL_C_WCHAR);
+
+        const char* pch = PyBytes_AS_STRING(query.Get());
+        SQLINTEGER  cch = (SQLINTEGER)(PyBytes_GET_SIZE(query.Get()) / (isWide ? sizeof(ODBCCHAR) : 1));
+
+        Py_BEGIN_ALLOW_THREADS
+        if (isWide)
+            ret = SQLExecDirectW(cur->hstmt, (SQLWCHAR*)pch, cch);
+        else
+            ret = SQLExecDirect(cur->hstmt, (SQLCHAR*)pch, cch);
+        Py_END_ALLOW_THREADS
     }
 
     if (cur->cnxn->hdbc == SQL_NULL_HANDLE)
@@ -701,7 +649,6 @@ static PyObject* execute(Cursor* cur, PyObject* pSql, PyObject* params, bool ski
         // The connection was closed by another thread in the ALLOW_THREADS block above.
 
         FreeParameterData(cur);
-
         return RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
     }
 
@@ -716,16 +663,18 @@ static PyObject* execute(Cursor* cur, PyObject* pSql, PyObject* params, bool ski
 
     while (ret == SQL_NEED_DATA)
     {
-        // We have bound a PyObject* using SQL_LEN_DATA_AT_EXEC, so ODBC is asking us for the data now.  We gave the
-        // PyObject pointer to ODBC in SQLBindParameter -- SQLParamData below gives the pointer back to us.
+        // One or more parameters were too long to bind normally so we set the
+        // length to SQL_LEN_DATA_AT_EXEC.  ODBC will return SQL_NEED_DATA for
+        // each of the parameters we did this for.
         //
-        // Note that we did not increment the pointer reference for this since we are still in the same C function call
-        // that performed the bind.
+        // For each one we set a pointer to the ParamInfo as the "parameter
+        // data" we can access with SQLParamData.  We've stashed everything we
+        // need in there.
 
         szLastFunction = "SQLParamData";
-        PyObject* pParam;
+        ParamInfo* pInfo;
         Py_BEGIN_ALLOW_THREADS
-        ret = SQLParamData(cur->hstmt, (SQLPOINTER*)&pParam);
+        ret = SQLParamData(cur->hstmt, (SQLPOINTER*)&pInfo);
         Py_END_ALLOW_THREADS
 
         if (ret != SQL_NEED_DATA && ret != SQL_NO_DATA && !SQL_SUCCEEDED(ret))
@@ -736,33 +685,15 @@ static PyObject* execute(Cursor* cur, PyObject* pSql, PyObject* params, bool ski
         if (ret == SQL_NEED_DATA)
         {
             szLastFunction = "SQLPutData";
-            if (PyUnicode_Check(pParam))
+            if (PyBytes_Check(pInfo->pObject))
             {
-                SQLWChar wchar(pParam); // Will convert to SQLWCHAR if necessary.
-
-                Py_ssize_t offset = 0;            // in characters
-                Py_ssize_t length = wchar.size(); // in characters
-
-                while (offset < length)
-                {
-                    SQLLEN remaining = min(cur->cnxn->varchar_maxlength, length - offset);
-                    Py_BEGIN_ALLOW_THREADS
-                    ret = SQLPutData(cur->hstmt, (SQLPOINTER)wchar[offset], (SQLLEN)(remaining * sizeof(ODBCCHAR)));
-                    Py_END_ALLOW_THREADS
-                    if (!SQL_SUCCEEDED(ret))
-                        return RaiseErrorFromHandle("SQLPutData", cur->cnxn->hdbc, cur->hstmt);
-                    offset += remaining;
-                }
-            }
-            else if (PyBytes_Check(pParam))
-            {
-                const char* p = PyBytes_AS_STRING(pParam);
+                const char* p = PyBytes_AS_STRING(pInfo->pObject);
                 SQLLEN offset = 0;
-                SQLLEN cb = (SQLLEN)PyBytes_GET_SIZE(pParam);
+                SQLLEN cb = (SQLLEN)PyBytes_GET_SIZE(pInfo->pObject);
                 while (offset < cb)
                 {
-                    SQLLEN remaining = min(cur->cnxn->varchar_maxlength, cb - offset);
-                    TRACE("SQLPutData [%d] (%d) %s\n", offset, remaining, &p[offset]);
+                    SQLLEN remaining = min(pInfo->maxlength, cb - offset);
+                    TRACE("SQLPutData [%d] (%d) %.10s\n", offset, remaining, &p[offset]);
                     Py_BEGIN_ALLOW_THREADS
                     ret = SQLPutData(cur->hstmt, (SQLPOINTER)&p[offset], remaining);
                     Py_END_ALLOW_THREADS
@@ -772,15 +703,15 @@ static PyObject* execute(Cursor* cur, PyObject* pSql, PyObject* params, bool ski
                 }
             }
 #if PY_VERSION_HEX >= 0x02060000
-            else if (PyByteArray_Check(pParam))
+            else if (PyByteArray_Check(pInfo->pObject))
             {
-                const char* p = PyByteArray_AS_STRING(pParam);
+                const char* p = PyByteArray_AS_STRING(pInfo->pObject);
                 SQLLEN offset = 0;
-                SQLLEN cb     = (SQLLEN)PyByteArray_GET_SIZE(pParam);
+                SQLLEN cb     = (SQLLEN)PyByteArray_GET_SIZE(pInfo->pObject);
                 while (offset < cb)
                 {
-                    SQLLEN remaining = min(cur->cnxn->varchar_maxlength, cb - offset);
-                    TRACE("SQLPutData [%d] (%d) %s\n", offset, remaining, &p[offset]);
+                    SQLLEN remaining = min(pInfo->maxlength, cb - offset);
+                    TRACE("SQLPutData [%d] (%d) %.10s\n", offset, remaining, &p[offset]);
                     Py_BEGIN_ALLOW_THREADS
                     ret = SQLPutData(cur->hstmt, (SQLPOINTER)&p[offset], remaining);
                     Py_END_ALLOW_THREADS
@@ -791,12 +722,12 @@ static PyObject* execute(Cursor* cur, PyObject* pSql, PyObject* params, bool ski
             }
 #endif
 #if PY_MAJOR_VERSION < 3
-            else if (PyBuffer_Check(pParam))
+            else if (PyBuffer_Check(pInfo->pObject))
             {
                 // Buffers can have multiple segments, so we might need multiple writes.  Looping through buffers isn't
                 // difficult, but we've wrapped it up in an iterator object to keep this loop simple.
 
-                BufferSegmentIterator it(pParam);
+                BufferSegmentIterator it(pInfo->pObject);
                 byte* pb;
                 SQLLEN cb;
                 while (it.Next(pb, cb))
@@ -1016,7 +947,7 @@ static PyObject* Cursor_executemany(PyObject* self, PyObject* args)
                 return 0;
             }
         }
-        
+
         if (PyErr_Occurred())
             return 0;
     }
@@ -1146,6 +1077,25 @@ static PyObject* Cursor_iternext(PyObject* self)
     return result;
 }
 
+static PyObject* Cursor_fetchval(PyObject* self, PyObject* args)
+{
+    UNUSED(args);
+
+    Cursor* cursor = Cursor_Validate(self, CURSOR_REQUIRE_RESULTS | CURSOR_RAISE_ERROR);
+    if (!cursor)
+        return 0;
+
+    Object row(Cursor_fetch(cursor));
+
+    if (!row)
+    {
+        if (PyErr_Occurred())
+            return 0;
+        Py_RETURN_NONE;
+    }
+
+    return Row_item(row, 0);
+}
 
 static PyObject* Cursor_fetchone(PyObject* self, PyObject* args)
 {
@@ -1233,7 +1183,7 @@ static PyObject* Cursor_tables(PyObject* self, PyObject* args, PyObject* kwargs)
         return 0;
 
     Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
-    
+
     if (!free_results(cur, FREE_STATEMENT | FREE_PREPARED))
         return 0;
 
@@ -1302,7 +1252,7 @@ static PyObject* Cursor_columns(PyObject* self, PyObject* args, PyObject* kwargs
         return 0;
 
     Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
-    
+
     if (!free_results(cur, FREE_STATEMENT | FREE_PREPARED))
         return 0;
 
@@ -1374,7 +1324,7 @@ static PyObject* Cursor_statistics(PyObject* self, PyObject* args, PyObject* kwa
         return 0;
 
     Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
-    
+
     if (!free_results(cur, FREE_STATEMENT | FREE_PREPARED))
         return 0;
 
@@ -1451,7 +1401,7 @@ static PyObject* _specialColumns(PyObject* self, PyObject* args, PyObject* kwarg
         return 0;
 
     Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
-    
+
     if (!free_results(cur, FREE_STATEMENT | FREE_PREPARED))
         return 0;
 
@@ -1522,7 +1472,7 @@ static PyObject* Cursor_primaryKeys(PyObject* self, PyObject* args, PyObject* kw
         return 0;
 
     Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
-    
+
     if (!free_results(cur, FREE_STATEMENT | FREE_PREPARED))
         return 0;
 
@@ -1593,7 +1543,7 @@ static PyObject* Cursor_foreignKeys(PyObject* self, PyObject* args, PyObject* kw
         return 0;
 
     Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
-    
+
     if (!free_results(cur, FREE_STATEMENT | FREE_PREPARED))
         return 0;
 
@@ -1655,20 +1605,20 @@ static PyObject* Cursor_getTypeInfo(PyObject* self, PyObject* args, PyObject* kw
 {
     UNUSED(kwargs);
 
-    SQLSMALLINT nDataType = SQL_ALL_TYPES;
+    int nDataType = SQL_ALL_TYPES;
 
     if (!PyArg_ParseTuple(args, "|i", &nDataType))
         return 0;
 
     Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
-    
+
     if (!free_results(cur, FREE_STATEMENT | FREE_PREPARED))
         return 0;
 
     SQLRETURN ret = 0;
 
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLGetTypeInfo(cur->hstmt, nDataType);
+    ret = SQLGetTypeInfo(cur->hstmt, (SQLSMALLINT)nDataType);
     Py_END_ALLOW_THREADS
 
     if (!SQL_SUCCEEDED(ret))
@@ -1713,6 +1663,37 @@ static PyObject* Cursor_nextset(PyObject* self, PyObject* args)
         free_results(cur, FREE_STATEMENT | KEEP_PREPARED);
         Py_RETURN_FALSE;
     }
+    if (!SQL_SUCCEEDED(ret))
+    {
+        TRACE("nextset: %d not SQL_SUCCEEDED\n", ret);
+        // Note: The SQL Server driver sometimes returns HY007 here if multiple statements (separated by ;) were
+        // submitted.  This is not documented, but I've seen it with multiple successful inserts.
+
+        PyObject* pError = GetErrorFromHandle("SQLMoreResults", cur->cnxn->hdbc, cur->hstmt);
+        //
+        // free_results must be run after the error has been collected
+        // from the cursor as it's lost otherwise.
+        // If free_results raises an error (eg a lost connection) report that instead.
+        //
+        if (!free_results(cur, FREE_STATEMENT | KEEP_PREPARED)) {
+            return 0;
+        }
+        //
+        // Return any error from the GetErrorFromHandle call above.
+        //
+        if (pError)
+        {
+            RaiseErrorFromException(pError);
+            Py_DECREF(pError);
+            return 0;
+        }
+
+        //
+        // Not clear how we'd get here, but if we're in an error state
+        // without an error, behave as if we had no nextset
+        //
+        Py_RETURN_FALSE;
+    }
 
     SQLSMALLINT cCols;
     Py_BEGIN_ALLOW_THREADS
@@ -1723,8 +1704,9 @@ static PyObject* Cursor_nextset(PyObject* self, PyObject* args)
         // Note: The SQL Server driver sometimes returns HY007 here if multiple statements (separated by ;) were
         // submitted.  This is not documented, but I've seen it with multiple successful inserts.
 
+        PyObject* pError = GetErrorFromHandle("SQLNumResultCols", cur->cnxn->hdbc, cur->hstmt);
         free_results(cur, FREE_STATEMENT | KEEP_PREPARED);
-        return RaiseErrorFromHandle("SQLNumResultCols", cur->cnxn->hdbc, cur->hstmt);
+        return pError;
     }
     free_results(cur, KEEP_STATEMENT | KEEP_PREPARED);
 
@@ -1788,7 +1770,7 @@ static PyObject* Cursor_procedureColumns(PyObject* self, PyObject* args, PyObjec
         return 0;
 
     Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
-    
+
     if (!free_results(cur, FREE_STATEMENT | FREE_PREPARED))
         return 0;
 
@@ -1847,7 +1829,7 @@ static PyObject* Cursor_procedures(PyObject* self, PyObject* args, PyObject* kwa
         return 0;
 
     Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
-    
+
     if (!free_results(cur, FREE_STATEMENT | FREE_PREPARED))
         return 0;
 
@@ -2060,6 +2042,12 @@ static char nextset_doc[] = "nextset() --> True | None\n" \
 
 static char ignored_doc[] = "Ignored.";
 
+static char fetchval_doc[] =
+    "fetchval() --> value | None\n" \
+    "\n"
+    "Returns the first column of the next row in the result set or None\n" \
+    "if there are no more rows.";
+
 static char fetchone_doc[] =
     "fetchone() --> Row | None\n" \
     "\n" \
@@ -2122,11 +2110,11 @@ static PyObject* Cursor_exit(PyObject* self, PyObject* args)
         if (!SQL_SUCCEEDED(ret))
             return RaiseErrorFromHandle("SQLEndTran(SQL_COMMIT)", cursor->cnxn->hdbc, cursor->hstmt);
     }
-    
+
     Py_RETURN_NONE;
 }
 
-    
+
 static PyMethodDef Cursor_methods[] =
 {
     { "close",            (PyCFunction)Cursor_close,            METH_NOARGS,                close_doc            },
@@ -2134,6 +2122,7 @@ static PyMethodDef Cursor_methods[] =
     { "executemany",      (PyCFunction)Cursor_executemany,      METH_VARARGS,               executemany_doc      },
     { "setinputsizes",    (PyCFunction)Cursor_ignored,          METH_VARARGS,               ignored_doc          },
     { "setoutputsize",    (PyCFunction)Cursor_ignored,          METH_VARARGS,               ignored_doc          },
+    { "fetchval",         (PyCFunction)Cursor_fetchval,         METH_NOARGS,                fetchval_doc         },
     { "fetchone",         (PyCFunction)Cursor_fetchone,         METH_NOARGS,                fetchone_doc         },
     { "fetchall",         (PyCFunction)Cursor_fetchall,         METH_NOARGS,                fetchall_doc         },
     { "fetchmany",        (PyCFunction)Cursor_fetchmany,        METH_VARARGS,               fetchmany_doc        },
@@ -2267,7 +2256,7 @@ Cursor_New(Connection* cnxn)
         if (cnxn->timeout)
         {
             Py_BEGIN_ALLOW_THREADS
-            ret = SQLSetStmtAttr(cur->hstmt, SQL_ATTR_QUERY_TIMEOUT, (SQLPOINTER)cnxn->timeout, 0);
+            ret = SQLSetStmtAttr(cur->hstmt, SQL_ATTR_QUERY_TIMEOUT, (SQLPOINTER)(uintptr_t)cnxn->timeout, 0);
             Py_END_ALLOW_THREADS
 
             if (!SQL_SUCCEEDED(ret))
@@ -2288,4 +2277,3 @@ void Cursor_init()
 {
     PyDateTime_IMPORT;
 }
-
