@@ -1,12 +1,21 @@
 
+// https://msdn.microsoft.com/en-us/library/ms711014(v=vs.85).aspx
+//
+// "The length of both the data buffer and the data it contains is measured in bytes, as
+// opposed to characters."
+//
+// https://msdn.microsoft.com/en-us/library/ms711786(v=vs.85).aspx
+//
+// Column Size: "For character types, this is the length in characters of the data"
 
 #include "pyodbc.h"
+#include "wrapper.h"
+#include "textenc.h"
 #include "pyodbcmodule.h"
 #include "params.h"
 #include "cursor.h"
 #include "connection.h"
 #include "buffer.h"
-#include "wrapper.h"
 #include "errors.h"
 #include "dbspecific.h"
 #include "sqlwchar.h"
@@ -26,90 +35,11 @@ static void FreeInfos(ParamInfo* a, Py_ssize_t count)
     {
         if (a[i].allocated)
             pyodbc_free(a[i].ParameterValuePtr);
-        Py_XDECREF(a[i].pParam);
+        Py_XDECREF(a[i].pObject);
     }
     pyodbc_free(a);
 }
 
-#define _MAKESTR(n) case n: return #n
-static const char* SqlTypeName(SQLSMALLINT n)
-{
-    switch (n)
-    {
-        _MAKESTR(SQL_UNKNOWN_TYPE);
-        _MAKESTR(SQL_CHAR);
-        _MAKESTR(SQL_VARCHAR);
-        _MAKESTR(SQL_LONGVARCHAR);
-        _MAKESTR(SQL_NUMERIC);
-        _MAKESTR(SQL_DECIMAL);
-        _MAKESTR(SQL_INTEGER);
-        _MAKESTR(SQL_SMALLINT);
-        _MAKESTR(SQL_FLOAT);
-        _MAKESTR(SQL_REAL);
-        _MAKESTR(SQL_DOUBLE);
-        _MAKESTR(SQL_DATETIME);
-        _MAKESTR(SQL_WCHAR);
-        _MAKESTR(SQL_WVARCHAR);
-        _MAKESTR(SQL_WLONGVARCHAR);
-        _MAKESTR(SQL_TYPE_DATE);
-        _MAKESTR(SQL_TYPE_TIME);
-        _MAKESTR(SQL_TYPE_TIMESTAMP);
-        _MAKESTR(SQL_SS_TIME2);
-        _MAKESTR(SQL_SS_XML);
-        _MAKESTR(SQL_BINARY);
-        _MAKESTR(SQL_VARBINARY);
-        _MAKESTR(SQL_LONGVARBINARY);
-    }
-    return "unknown";
-}
-
-static const char* CTypeName(SQLSMALLINT n)
-{
-    switch (n)
-    {
-        _MAKESTR(SQL_C_CHAR);
-        _MAKESTR(SQL_C_WCHAR);
-        _MAKESTR(SQL_C_LONG);
-        _MAKESTR(SQL_C_SHORT);
-        _MAKESTR(SQL_C_FLOAT);
-        _MAKESTR(SQL_C_DOUBLE);
-        _MAKESTR(SQL_C_NUMERIC);
-        _MAKESTR(SQL_C_DEFAULT);
-        _MAKESTR(SQL_C_DATE);
-        _MAKESTR(SQL_C_TIME);
-        _MAKESTR(SQL_C_TIMESTAMP);
-        _MAKESTR(SQL_C_TYPE_DATE);
-        _MAKESTR(SQL_C_TYPE_TIME);
-        _MAKESTR(SQL_C_TYPE_TIMESTAMP);
-        _MAKESTR(SQL_C_INTERVAL_YEAR);
-        _MAKESTR(SQL_C_INTERVAL_MONTH);
-        _MAKESTR(SQL_C_INTERVAL_DAY);
-        _MAKESTR(SQL_C_INTERVAL_HOUR);
-        _MAKESTR(SQL_C_INTERVAL_MINUTE);
-        _MAKESTR(SQL_C_INTERVAL_SECOND);
-        _MAKESTR(SQL_C_INTERVAL_YEAR_TO_MONTH);
-        _MAKESTR(SQL_C_INTERVAL_DAY_TO_HOUR);
-        _MAKESTR(SQL_C_INTERVAL_DAY_TO_MINUTE);
-        _MAKESTR(SQL_C_INTERVAL_DAY_TO_SECOND);
-        _MAKESTR(SQL_C_INTERVAL_HOUR_TO_MINUTE);
-        _MAKESTR(SQL_C_INTERVAL_HOUR_TO_SECOND);
-        _MAKESTR(SQL_C_INTERVAL_MINUTE_TO_SECOND);
-        _MAKESTR(SQL_C_BINARY);
-        _MAKESTR(SQL_C_BIT);
-        _MAKESTR(SQL_C_SBIGINT);
-        _MAKESTR(SQL_C_UBIGINT);
-        _MAKESTR(SQL_C_TINYINT);
-        _MAKESTR(SQL_C_SLONG);
-        _MAKESTR(SQL_C_SSHORT);
-        _MAKESTR(SQL_C_STINYINT);
-        _MAKESTR(SQL_C_ULONG);
-        _MAKESTR(SQL_C_USHORT);
-        _MAKESTR(SQL_C_UTINYINT);
-        _MAKESTR(SQL_C_GUID);
-    }
-    return "unknown";
-}
-
 static bool GetNullInfo(Cursor* cur, Py_ssize_t index, ParamInfo& info)
 {
     if (!GetParamType(cur, index, info.ParameterType))
@@ -132,92 +62,142 @@ static bool GetNullBinaryInfo(Cursor* cur, Py_ssize_t index, ParamInfo& info)
 }
 
 
+#if PY_MAJOR_VERSION >= 3
 static bool GetBytesInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamInfo& info)
 {
-    // In Python 2, a bytes object (ANSI string) is passed as varchar.  In Python 3, it is passed as binary.
+    // The Python 3 version that writes bytes as binary data.
+    Py_ssize_t cb = PyBytes_GET_SIZE(param);
 
-    Py_ssize_t len = PyBytes_GET_SIZE(param);
+    info.ValueType  = SQL_C_BINARY;
+    info.ColumnSize = (SQLUINTEGER)max(cb, 1);
 
-#if PY_MAJOR_VERSION >= 3
-    info.ValueType = SQL_C_BINARY;
-    info.ColumnSize = (SQLUINTEGER)max(len, 1);
+    SQLLEN maxlength = cur->cnxn->GetMaxLength(info.ValueType);
 
-    if (len <= cur->cnxn->binary_maxlength)
+    if (maxlength == 0 || cb <= maxlength)
     {
         info.ParameterType     = SQL_VARBINARY;
-        info.StrLen_or_Ind     = len;
+        info.StrLen_or_Ind     = cb;
+        info.BufferLength      = (SQLLEN)info.ColumnSize;
         info.ParameterValuePtr = PyBytes_AS_STRING(param);
     }
     else
     {
         // Too long to pass all at once, so we'll provide the data at execute.
         info.ParameterType     = SQL_LONGVARBINARY;
-        info.StrLen_or_Ind     = cur->cnxn->need_long_data_len ? SQL_LEN_DATA_AT_EXEC((SQLLEN)len) : SQL_DATA_AT_EXEC;
-        info.ParameterValuePtr = param;
+        info.StrLen_or_Ind     = cur->cnxn->need_long_data_len ? SQL_LEN_DATA_AT_EXEC((SQLLEN)cb) : SQL_DATA_AT_EXEC;
+        info.ParameterValuePtr = &info;
+        info.BufferLength      = sizeof(ParamInfo*);
+        info.pObject           = param;
+        Py_INCREF(info.pObject);
+        info.maxlength = maxlength;
     }
 
-#else
-    info.ValueType = SQL_C_CHAR;
-    info.ColumnSize = (SQLUINTEGER)max(len, 1);
-
-    if (len <= cur->cnxn->varchar_maxlength)
-    {
-        info.ParameterType     = SQL_VARCHAR;
-        info.StrLen_or_Ind     = len;
-        info.ParameterValuePtr = PyBytes_AS_STRING(param);
-    }
-    else
-    {
-        // Too long to pass all at once, so we'll provide the data at execute.
-        info.ParameterType     = SQL_LONGVARCHAR;
-        info.StrLen_or_Ind     = cur->cnxn->need_long_data_len ? SQL_LEN_DATA_AT_EXEC((SQLLEN)len) : SQL_DATA_AT_EXEC;
-        info.ParameterValuePtr = param;
-    }
-#endif
-
     return true;
 }
+#endif
 
-static bool GetUnicodeInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamInfo& info)
+#if PY_MAJOR_VERSION < 3
+static bool GetStrInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamInfo& info)
 {
-    Py_UNICODE* pch = PyUnicode_AsUnicode(param);
-    Py_ssize_t  len = PyUnicode_GET_SIZE(param);
+    const TextEnc& enc = cur->cnxn->str_enc;
 
-    info.ValueType  = SQL_C_WCHAR;
-    info.ColumnSize = (SQLUINTEGER)max(len, 1);
+    info.ValueType = enc.ctype;
 
-    if (len <= cur->cnxn->wvarchar_maxlength)
+    Py_ssize_t cch = PyString_GET_SIZE(param);
+
+    info.ColumnSize = (SQLUINTEGER)max(cch, 1);
+
+    Object encoded;
+
+    if (enc.optenc == OPTENC_RAW)
     {
-        if (ODBCCHAR_SIZE == Py_UNICODE_SIZE)
-        {
-            info.ParameterValuePtr = pch;
-        }
-        else
-        {
-            // SQLWCHAR and Py_UNICODE are not the same size, so we need to allocate and copy a buffer.
-            if (len > 0)
-            {
-                info.ParameterValuePtr = SQLWCHAR_FromUnicode(pch, len);
-                if (info.ParameterValuePtr == 0)
-                    return false;
-                info.allocated = true;
-            }
-            else
-            {
-                info.ParameterValuePtr = pch;
-            }
-        }
+        // Take the text as-is.  This is not really a good idea since users will need to make
+        // sure the encoding over the wire matches their system encoding, but it will be wanted
+        // and it is fast when you get it to work.
+        encoded = param;
+    }
+    else
+    {
+        // Encode the text with the user's encoding.
+        encoded = PyCodec_Encode(param, enc.name, "strict");
+        if (!encoded)
+            return false;
 
-        info.ParameterType = SQL_WVARCHAR;
-        info.StrLen_or_Ind = (SQLINTEGER)(len * ODBCCHAR_SIZE);
+        if (!PyBytes_CheckExact(encoded))
+        {
+            // Not all encodings return bytes.
+            PyErr_Format(PyExc_TypeError, "Unicode read encoding '%s' returned unexpected data type: %s",
+                         enc.name, encoded.Get()->ob_type->tp_name);
+            return false;
+        }
+    }
+
+    Py_ssize_t cb = PyBytes_GET_SIZE(encoded);
+    info.pObject = encoded.Detach();
+
+    SQLLEN maxlength = cur->cnxn->GetMaxLength(info.ValueType);
+    if (maxlength == 0 || cb <= maxlength)
+    {
+        info.ParameterType     = (enc.ctype == SQL_C_CHAR) ? SQL_VARCHAR : SQL_WVARCHAR;
+        info.ParameterValuePtr = PyBytes_AS_STRING(info.pObject);
+        info.StrLen_or_Ind     = (SQLINTEGER)cb;
     }
     else
     {
         // Too long to pass all at once, so we'll provide the data at execute.
+        info.ParameterType     = (enc.ctype == SQL_C_CHAR) ? SQL_LONGVARCHAR : SQL_WLONGVARCHAR;
+        info.ParameterValuePtr = &info;
+        info.BufferLength      = sizeof(ParamInfo*);
+        info.StrLen_or_Ind     = cur->cnxn->need_long_data_len ? SQL_LEN_DATA_AT_EXEC((SQLINTEGER)cb) : SQL_DATA_AT_EXEC;
+        info.maxlength = maxlength;
+    }
 
-        info.ParameterType     = SQL_WLONGVARCHAR;
-        info.StrLen_or_Ind     = cur->cnxn->need_long_data_len ? SQL_LEN_DATA_AT_EXEC((SQLLEN)len * ODBCCHAR_SIZE) : SQL_DATA_AT_EXEC;
-        info.ParameterValuePtr = param;
+    return true;
+}
+#endif
+
+
+static bool GetUnicodeInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamInfo& info)
+{
+    const TextEnc& enc = cur->cnxn->unicode_enc;
+
+    info.ValueType = enc.ctype;
+
+    Py_ssize_t cch = PyUnicode_GET_SIZE(param);
+
+    info.ColumnSize = (SQLUINTEGER)max(cch, 1);
+
+    Object encoded(PyCodec_Encode(param, enc.name, "strict"));
+    if (!encoded)
+        return false;
+
+    if (enc.optenc == OPTENC_NONE && !PyBytes_CheckExact(encoded))
+    {
+        PyErr_Format(PyExc_TypeError, "Unicode write encoding '%s' returned unexpected data type: %s",
+                     enc.name, encoded.Get()->ob_type->tp_name);
+        return false;
+    }
+
+    Py_ssize_t cb = PyBytes_GET_SIZE(encoded);
+    info.pObject = encoded.Detach();
+
+    SQLLEN maxlength = cur->cnxn->GetMaxLength(enc.ctype);
+
+    if (maxlength == 0 || cb <= maxlength)
+    {
+        info.ParameterType     = (enc.ctype == SQL_C_CHAR) ? SQL_VARCHAR : SQL_WVARCHAR;
+        info.ParameterValuePtr = PyBytes_AS_STRING(info.pObject);
+        info.BufferLength      = (SQLINTEGER)cb;
+        info.StrLen_or_Ind     = (SQLINTEGER)cb;
+    }
+    else
+    {
+        // Too long to pass all at once, so we'll provide the data at execute.
+        info.ParameterType     = (enc.ctype == SQL_C_CHAR) ? SQL_LONGVARCHAR : SQL_WLONGVARCHAR;
+        info.ParameterValuePtr = &info;
+        info.BufferLength      = sizeof(ParamInfo*);
+        info.StrLen_or_Ind     = cur->cnxn->need_long_data_len ? SQL_LEN_DATA_AT_EXEC((SQLINTEGER)cb) : SQL_DATA_AT_EXEC;
+        info.maxlength = maxlength;
     }
 
     return true;
@@ -319,13 +299,38 @@ static bool GetIntInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamInfo
 
 static bool GetLongInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamInfo& info)
 {
-    // TODO: Overflow?
-    info.Data.i64 = (INT64)PyLong_AsLongLong(param);
+    // Try to use integer when possible.  BIGINT is not always supported and is a "special
+    // case" for some drivers.
 
-    info.ValueType         = SQL_C_SBIGINT;
-    info.ParameterType     = SQL_BIGINT;
-    info.ParameterValuePtr = &info.Data.i64;
-    return true;
+    // REVIEW: C & C++ now have constants for max sizes, but I'm not sure if they will be
+    // available on all platforms Python supports right now.  It would be more performant when
+    // a lot of 64-bit values are used since we could avoid the AsLong call.
+
+    // It's not clear what the maximum SQL_INTEGER should be.  The Microsoft documentation says
+    // it is a 'long int', but some drivers run into trouble at high values.  We'll use
+    // SQL_INTEGER as an optimization for smaller values and rely on BIGINT
+
+    int overflow = 0;
+    info.Data.l = PyLong_AsLongAndOverflow(param, &overflow);
+    if (overflow == 0 && !PyErr_Occurred() && (info.Data.l <= 0x7FFFFFFF))
+    {
+        info.ValueType         = SQL_C_LONG;
+        info.ParameterType     = SQL_INTEGER;
+        info.ParameterValuePtr = &info.Data.l;
+    }
+    else
+    {
+        PyErr_Clear();
+        info.Data.i64 = (INT64)PyLong_AsLongLong(param);
+        if (!PyErr_Occurred())
+        {
+            info.ValueType         = SQL_C_SBIGINT;
+            info.ParameterType     = SQL_BIGINT;
+            info.ParameterValuePtr = &info.Data.i64;
+        }
+    }
+
+    return !PyErr_Occurred();
 }
 
 static bool GetFloatInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamInfo& info)
@@ -416,13 +421,42 @@ static char* CreateDecimalString(long sign, PyObject* digits, long exp)
     return pch;
 }
 
-static bool GetDecimalInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamInfo& info)
+static bool GetUUIDInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamInfo& info, PyObject* uuid_type)
 {
+    // uuid_type: This is a new reference that we are responsible for freeing.
+    Object tmp(uuid_type);
+
+    info.ValueType = SQL_C_GUID;
+    info.ParameterType = SQL_GUID;
+    info.ColumnSize = 16;
+
+    info.allocated = true;
+    info.ParameterValuePtr = pyodbc_malloc(sizeof(SQLGUID));
+    if (!info.ParameterValuePtr)
+    {
+        PyErr_NoMemory();
+        return false;
+    }
+
+    // Do we need to use "bytes" on a big endian machine?
+    Object b(PyObject_GetAttrString(param, "bytes_le"));
+    if (!b)
+        return false;
+    memcpy(info.ParameterValuePtr, PyBytes_AS_STRING(b.Get()), sizeof(SQLGUID));
+    return true;
+}
+
+
+static bool GetDecimalInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamInfo& info, PyObject* decimal_type)
+{
+    // decimal_type: This is a new reference that we are responsible for freeing.
+    Object tmp(decimal_type);
+
     // The NUMERIC structure never works right with SQL Server and probably a lot of other drivers.  We'll bind as a
     // string.  Unfortunately, the Decimal class doesn't seem to have a way to force it to return a string without
     // exponents, so we'll have to build it ourselves.
 
-    Object t = PyObject_CallMethod(param, "as_tuple", 0);
+    Object t(PyObject_CallMethod(param, "as_tuple", 0));
     if (!t)
         return false;
 
@@ -479,15 +513,16 @@ static bool GetBufferInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamI
     const char* pb;
     Py_ssize_t  cb = PyBuffer_GetMemory(param, &pb);
 
-    if (cb != -1 && cb <= cur->cnxn->binary_maxlength)
+    SQLLEN maxlength = cur->cnxn->GetMaxLength(info.ValueType);
+    if (maxlength == 0 || cb <= maxlength)
     {
         // There is one segment, so we can bind directly into the buffer object.
 
         info.ParameterType     = SQL_VARBINARY;
         info.ParameterValuePtr = (SQLPOINTER)pb;
-        info.BufferLength      = cb;
+        info.BufferLength      = (SQLINTEGER)cb;
         info.ColumnSize        = (SQLUINTEGER)max(cb, 1);
-        info.StrLen_or_Ind     = cb;
+        info.StrLen_or_Ind     = (SQLINTEGER)cb;
     }
     else
     {
@@ -496,10 +531,13 @@ static bool GetBufferInfo(Cursor* cur, Py_ssize_t index, PyObject* param, ParamI
         // need to up the refcount!)
 
         info.ParameterType     = SQL_LONGVARBINARY;
-        info.ParameterValuePtr = param;
+        info.ParameterValuePtr = &info;
+        info.BufferLength      = sizeof(ParamInfo*);
         info.ColumnSize        = (SQLUINTEGER)PyBuffer_Size(param);
-        info.BufferLength      = sizeof(PyObject*); // How big is ParameterValuePtr; ODBC copies it and gives it back in SQLParamData
         info.StrLen_or_Ind     = cur->cnxn->need_long_data_len ? SQL_LEN_DATA_AT_EXEC((SQLLEN)PyBuffer_Size(param)) : SQL_DATA_AT_EXEC;
+        info.pObject = param;
+        Py_INCREF(info.pObject);
+        info.maxlength = maxlength;
     }
 
     return true;
@@ -512,21 +550,26 @@ static bool GetByteArrayInfo(Cursor* cur, Py_ssize_t index, PyObject* param, Par
     info.ValueType = SQL_C_BINARY;
 
     Py_ssize_t cb = PyByteArray_Size(param);
-    if (cb <= cur->cnxn->binary_maxlength)
+
+    SQLLEN maxlength = cur->cnxn->GetMaxLength(info.ValueType);
+    if (maxlength == 0 || cb <= maxlength)
     {
         info.ParameterType     = SQL_VARBINARY;
         info.ParameterValuePtr = (SQLPOINTER)PyByteArray_AsString(param);
-        info.BufferLength      = cb;
+        info.BufferLength      = (SQLINTEGER)cb;
         info.ColumnSize        = (SQLUINTEGER)max(cb, 1);
-        info.StrLen_or_Ind     = cb;
+        info.StrLen_or_Ind     = (SQLINTEGER)cb;
     }
     else
     {
         info.ParameterType     = SQL_LONGVARBINARY;
-        info.ParameterValuePtr = param;
+        info.ParameterValuePtr = &info;
+        info.BufferLength      = sizeof(ParamInfo*);
         info.ColumnSize        = (SQLUINTEGER)cb;
-        info.BufferLength      = sizeof(PyObject*); // How big is ParameterValuePtr; ODBC copies it and gives it back in SQLParamData
         info.StrLen_or_Ind     = cur->cnxn->need_long_data_len ? SQL_LEN_DATA_AT_EXEC((SQLLEN)cb) : SQL_DATA_AT_EXEC;
+        info.pObject = param;
+        Py_INCREF(info.pObject);
+        info.maxlength = maxlength;
     }
     return true;
 }
@@ -538,17 +581,19 @@ static bool GetParameterInfo(Cursor* cur, Py_ssize_t index, PyObject* param, Par
     //
     // Populates `info`.
 
-    // Hold a reference to param until info is freed, because info will often be holding data borrowed from param.
-    info.pParam = param;
-
     if (param == Py_None)
         return GetNullInfo(cur, index, info);
 
     if (param == null_binary)
         return GetNullBinaryInfo(cur, index, info);
 
+#if PY_MAJOR_VERSION >= 3
     if (PyBytes_Check(param))
         return GetBytesInfo(cur, index, param, info);
+#else
+    if (PyBytes_Check(param))
+        return GetStrInfo(cur, index, param, info);
+#endif
 
     if (PyUnicode_Check(param))
         return GetUnicodeInfo(cur, index, param, info);
@@ -571,9 +616,6 @@ static bool GetParameterInfo(Cursor* cur, Py_ssize_t index, PyObject* param, Par
     if (PyFloat_Check(param))
         return GetFloatInfo(cur, index, param, info);
 
-    if (PyDecimal_Check(param))
-        return GetDecimalInfo(cur, index, param, info);
-
 #if PY_VERSION_HEX >= 0x02060000
     if (PyByteArray_Check(param))
         return GetByteArrayInfo(cur, index, param, info);
@@ -587,13 +629,30 @@ static bool GetParameterInfo(Cursor* cur, Py_ssize_t index, PyObject* param, Par
         return GetBufferInfo(cur, index, param, info);
 #endif
 
+    // Decimal
+
+    PyObject* cls = 0;
+    if (!IsInstanceForThread(param, "decimal", "Decimal", &cls))
+        return false;
+
+    if (cls != 0)
+        return GetDecimalInfo(cur, index, param, info, cls);
+
+    // UUID
+
+    if (!IsInstanceForThread(param, "uuid", "UUID", &cls))
+        return false;
+
+    if (cls != 0)
+        return GetUUIDInfo(cur, index, param, info, cls);
+
     RaiseErrorV("HY105", ProgrammingError, "Invalid parameter type.  param-index=%zd param-type=%s", index, Py_TYPE(param)->tp_name);
     return false;
 }
 
 bool BindParameter(Cursor* cur, Py_ssize_t index, ParamInfo& info)
 {
-    TRACE("BIND: param=%d ValueType=%d (%s) ParameterType=%d (%s) ColumnSize=%d DecimalDigits=%d BufferLength=%d *pcb=%d\n",
+    TRACE("BIND: param=%ld ValueType=%d (%s) ParameterType=%d (%s) ColumnSize=%ld DecimalDigits=%d BufferLength=%ld *pcb=%ld\n",
           (index+1), info.ValueType, CTypeName(info.ValueType), info.ParameterType, SqlTypeName(info.ParameterType), info.ColumnSize,
           info.DecimalDigits, info.BufferLength, info.StrLen_or_Ind);
 
@@ -682,32 +741,41 @@ bool PrepareAndBind(Cursor* cur, PyObject* pSql, PyObject* original_params, bool
         SQLSMALLINT cParamsT = 0;
         const char* szErrorFunc = "SQLPrepare";
 
-        if (PyUnicode_Check(pSql))
-        {
-            SQLWChar sql(pSql);
-            Py_BEGIN_ALLOW_THREADS
-            ret = SQLPrepareW(cur->hstmt, sql.get(), SQL_NTS);
-            if (SQL_SUCCEEDED(ret))
-            {
-                szErrorFunc = "SQLNumParams";
-                ret = SQLNumParams(cur->hstmt, &cParamsT);
-            }
-            Py_END_ALLOW_THREADS
-        }
+        const TextEnc* penc;
+
 #if PY_MAJOR_VERSION < 3
-        else
+        if (PyBytes_Check(pSql))
         {
-            TRACE("SQLPrepare(%s)\n", PyString_AS_STRING(pSql));
-            Py_BEGIN_ALLOW_THREADS
-            ret = SQLPrepare(cur->hstmt, (SQLCHAR*)PyString_AS_STRING(pSql), SQL_NTS);
-            if (SQL_SUCCEEDED(ret))
-            {
-                szErrorFunc = "SQLNumParams";
-                ret = SQLNumParams(cur->hstmt, &cParamsT);
-            }
-            Py_END_ALLOW_THREADS
+            penc = &cur->cnxn->str_enc;
         }
+        else
 #endif
+        {
+            penc = &cur->cnxn->unicode_enc;
+        }
+
+        Object query(penc->Encode(pSql));
+        if (!query)
+            return 0;
+
+        bool isWide = (penc->ctype == SQL_C_WCHAR);
+
+        const char* pch = PyBytes_AS_STRING(query.Get());
+        SQLINTEGER  cch = (SQLINTEGER)(PyBytes_GET_SIZE(query.Get()) / (isWide ? sizeof(ODBCCHAR) : 1));
+
+        TRACE("SQLPrepare(%s)\n", pch);
+
+        Py_BEGIN_ALLOW_THREADS
+        if (isWide)
+            ret = SQLPrepareW(cur->hstmt, (SQLWCHAR*)pch, cch);
+        else
+            ret = SQLPrepare(cur->hstmt, (SQLCHAR*)pch, cch);
+        if (SQL_SUCCEEDED(ret))
+        {
+            szErrorFunc = "SQLNumParams";
+            ret = SQLNumParams(cur->hstmt, &cParamsT);
+        }
+        Py_END_ALLOW_THREADS
 
         if (cur->cnxn->hdbc == SQL_NULL_HANDLE)
         {
@@ -748,10 +816,7 @@ bool PrepareAndBind(Cursor* cur, PyObject* pSql, PyObject* original_params, bool
 
     for (Py_ssize_t i = 0; i < cParams; i++)
     {
-        // PySequence_GetItem returns a *new* reference, which GetParameterInfo will take ownership of.  It is stored
-        // in paramInfos and will be released in FreeInfos (which is always eventually called).
-
-        PyObject* param = PySequence_GetItem(original_params, i + params_offset);
+        Object param(PySequence_GetItem(original_params, i + params_offset));
         if (!GetParameterInfo(cur, i, param, cur->paramInfos[i]))
         {
             FreeInfos(cur->paramInfos, cParams);
@@ -870,4 +935,3 @@ bool Params_init()
 
     return true;
 }
-
