@@ -13,9 +13,12 @@
 #define PYODBC_H
 
 #ifdef _MSC_VER
+// The MS headers generate a ton of warnings.
+#pragma warning(push, 0)
 #define _CRT_SECURE_NO_WARNINGS
 #include <windows.h>
 #include <malloc.h>
+#pragma warning(pop)
 typedef __int64 INT64;
 typedef unsigned __int64 UINT64;
 #else
@@ -24,6 +27,7 @@ typedef unsigned int UINT;
 typedef long long INT64;
 typedef unsigned long long UINT64;
 #define _strcmpi strcasecmp
+#define _strdup strdup
 #ifdef __MINGW32__
   #include <windef.h>
   #include <malloc.h>
@@ -105,7 +109,7 @@ inline void _strlwr(char* name)
   #include <crtdbg.h>
   inline void FailAssert(const char* szFile, size_t line, const char* szExpr)
   {
-      printf("assertion failed: %s(%d)\n%s\n", szFile, line, szExpr);
+      printf("assertion failed: %s(%d)\n%s\n", szFile, (int)line, szExpr);
       __debugbreak(); // _CrtDbgBreak();
   }
   #define I(expr) if (!(expr)) FailAssert(__FILE__, __LINE__, #expr);
@@ -122,17 +126,19 @@ inline void DebugTrace(const char* szFmt, ...) { UNUSED(szFmt); }
 #endif
 #define TRACE DebugTrace
 
-#ifdef PYODBC_LEAK_CHECK
-#define pyodbc_malloc(len) _pyodbc_malloc(__FILE__, __LINE__, len)
-void* _pyodbc_malloc(const char* filename, int lineno, size_t len);
-void pyodbc_free(void* p);
-void pyodbc_leak_check();
-#else
+// #ifdef PYODBC_LEAK_CHECK
+// #define pyodbc_malloc(len) _pyodbc_malloc(__FILE__, __LINE__, len)
+// void* _pyodbc_malloc(const char* filename, int lineno, size_t len);
+// void pyodbc_free(void* p);
+// void pyodbc_leak_check();
+// #else
 #define pyodbc_malloc malloc
 #define pyodbc_free free
-#endif
+// #endif
 
 void PrintBytes(void* p, size_t len);
+const char* CTypeName(SQLSMALLINT n);
+const char* SqlTypeName(SQLSMALLINT n);
 
 #include "pyodbccompat.h"
 
